#include "camera.h"
#include "nawiglcanvas.h"
#include <wx/glcanvas.h>
#include <float.h>

#define MAX(X,Y) ((X)>(Y)?(X):(Y))

double Camera::minimumDistanceToGround = 10000.;

Camera::Camera( double radius, NawiGLCanvas *parentCanvas )
	:
	worldRadius( radius ),
	parentCanvas(parentCanvas),
	referenceCenter(0.,0.,0.)
{
	reset(); // sets FoV, orientation, position
}

void Camera::setPosition( const Vector &v )
{
	Vector sphPos = v.cartesianToSpherical();

	altitude = sphPos.radius - worldRadius;
	if( altitude >= minimumDistanceToGround  )
	{
		if( ! wxIsNaN( sphPos.latitude ) )
			latitude.radians = sphPos.latitude;
		if( ! wxIsNaN( sphPos.longitude ) )
			longitude.radians = sphPos.longitude;

		position = v;
	}
}

// When resizing the window, the focal length is scaled.
// When resetting the camera (resetting==true), the focal length is
// reinitialized to the diagonal of the window.
void Camera::setViewport(const wxSize &screen, bool resetting)
{
	if( screen.x && screen.y )
	{
		viewport.width = screen.x;
		viewport.height = screen.y;
		glViewport(
			viewport.x,
			viewport.y,
			viewport.width,
			viewport.height
			);

		if( resetting )
			// Set a new normal focal length
			focal = Focal( viewport.getDiagonal() );
		else
			// Scale the existing focal length
			focal.update( viewport.getDiagonal() );

		updateProjection();
	}
}

void Camera::setView()
{
	viewMatrix.setRigidBodyTransformation(orientation, position - referenceCenter );

	// Variables position and orientation refer to the position and
	// orientation of the eye in world coordinates.
	// The MODELVIEW matrix refers to the position and orientation
	// of the world in eye coordinates.
	// So the matrix has to be inverted.
	viewMatrix.invertRigidBodyTransformation();

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixd(viewMatrix.val);
}

void Camera::update()
{
	setProjection(viewport);
	setView();
}

void Camera::reset()
{
	orientation = Quaternion( -0.5, -0.5, -0.5, 0.5 );
	setPosition( Vector( worldRadius * 2, 0., 0. ) );

	setViewport( parentCanvas->GetSize(), true /*resetting*/ );
}

void Camera::setProjection(Viewport &viewport)
{
	double aspectRatio =  viewport.getAspectRatio();
	double zNear = altitude / 10.;

	// The farthest point that can be viewed on a sphere using a
	// perspective projection is given by a line passing through the
	// camera position and tangent to the sphere.
	// The point of tangency, the center of the sphere and the camera
	// position form a right triangle whose right angle is at the point
	// of tangency.
	double cameraToCenter = altitude + worldRadius; // hypotenuse
	double zFar = sqrt(
			cameraToCenter*cameraToCenter
			- worldRadius*worldRadius
			); // tangent line

	projectionMatrix.setPerspectiveProjection(cotanHalfVerticalAoV, aspectRatio, zNear, zFar);

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixd(projectionMatrix.val);
}

//! Called when focal length is changed
void Camera::updateProjection()
{
	// To allow setting the perspective projection in the same way as with
	// a standard "physical" camera, this camera class uses window
	// dimensions in pixels as sensor dimensions.
	// The focal length, in pixels as well, is first initialized
	// to the diagonal of the window.
	//
	// The following formula refers to a right triangle defined by:
	// - the focal point
	// - the center of the sensor (right angle)
	// - a point on the limit of the sensor in a vertical plane
	//
	// The angle located at the focal point is the half vertical
	// angle of view. (halfVerticalAoV)
	// The projection matrix uses the cotangent of this angle, so:
	//
	// tan( halfVerticalAoV ) = opposite edge / adjacent edge
	// 			  = sensorHeight/2 / focalLength
	// cotan( halfVerticalAoV ) = 1 / tan( halfVerticalAoV )
	// 			    = 2. * focalLength / sensorHeight
	//
	// sensorHeight is the vertical field of view.
	// See http://en.wikipedia.org/wiki/Field_of_view_(image_processing)
	// and http://wiki.panotools.org/Field_of_View
	// if you need more information.

	cotanHalfVerticalAoV = 2. * focal.length / viewport.height;
	parentCanvas->Refresh();
}
