#include "texture.h"
#include <wx/string.h>
#include <wx/image.h>
#include <wx/glcanvas.h>
#include <wx/log.h>

//! Loads the texture from file to RAM
Texture::Texture(const wxString &fileName)
	:
	id(0),
	data(NULL),
	cantBeLoaded(false)
{
	wxLogDebug("Loading texture %s", fileName.c_str());

	// Don't forget to initialize image handlers in wxApp-derived class
	if(!wxFileExists(fileName))
	{
		cantBeLoaded = true;
		wxLogDebug("failed");
		return;
	}

	wxImage *img = new wxImage(fileName);

	width = img->GetWidth();
	height = img->GetHeight();
	if(!width || !height)
	{
		delete img;
		cantBeLoaded = true;
		wxLogDebug("failed");
		return;
	}
	float logWidth = log(width) / log(2.);
	float logHeight = log(height) / log(2.);
	if(logWidth != (int)logWidth || logHeight != (int)logHeight)
	{
		// Texture dimensions are not power of 2
		delete img;
		cantBeLoaded = true;
		wxLogDebug("failed");
		return;
	}

	// wxImage data will be copied to (unsigned char*)data, because:
	// - wxImage returns separate RGB and Alpha data
	// - OpenGL wants RGBA data
	// - wxImage stores pixels line by line, top-to-bottom, left-to-right
	// - OpenGL stores pixels line by line, bottom-to-top, left-to-right
	unsigned char *rgbData = img->GetData();
	unsigned char *alphaData = img->GetAlpha();

	bytesPerPixel = img->HasAlpha() ?  4 : 3;

	data = new unsigned char[ width * height * bytesPerPixel ];

	int lastRow = height - 1,
	    sourceIndex,
	    destIndex;

	for( int y = 0; y < height; y++ )
	{
		for( int x = 0; x < width; x++ )
		{
			sourceIndex = ( x + (lastRow-y) * width ) * 3;
			destIndex = ( x + y * width ) * bytesPerPixel;

			for( int i = 0; i < 3; i++ )
				data[destIndex + i] = rgbData[sourceIndex + i];

			if( bytesPerPixel == 4 )
				data[destIndex + 3]=alphaData[sourceIndex / 3];
		}
	}
	delete img;
}

/*! Deletes the texture. It should have already been removed from OpenGL memory
 *  using the unLoad() function. This is because the update thread cannot call
 *  unLoad(), which calls an OpenGL function.
 */
Texture::~Texture()
{
	if(data)
		delete[] data;
}

/*! Loads the texture from RAM to OpenGL
 *  This should be called by getId() when handling a Paint event, or by another
 *  function when handling an Idle event.
 *  Only the main thread can call this function.
 */
void Texture::load()
{
	if( !data )
		return;
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	// Disable mipmapping to allow using only one glTexImage2D
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D,
		0,
		bytesPerPixel,
		width,
		height,
		0,
		(bytesPerPixel == 4) ?  GL_RGBA : GL_RGB,
		GL_UNSIGNED_BYTE,
		data);
/*	gluBuild2DMipmaps(GL_TEXTURE_2D,
		bytesPerPixel,
		width,
		height,
		(bytesPerPixel == 4) ?  GL_RGBA : GL_RGB,
		GL_UNSIGNED_BYTE,
		data);*/
	delete[] data;
	data = NULL;
}

//! Removes the texture from OpenGL memory.
void Texture::unLoad()
{
	if(id)
		glDeleteTextures(1, &id);
}

//! Returns the OpenGL texture identifier, and loads the texture
//  if it is not already done.
unsigned int Texture::getId()
{
	if( id == 0 )
		load();
	return id;
}

//! Checks if the texture is loaded in OpenGL memory.
//  If it is not loaded, it can be deleted by the update thread.
bool Texture::isLoaded()
{
	return ( id != 0 );
}
