#include "vertextable.h"
#include "vector.h"
#include "color.h"
#include "texture.h"
#define GL_GLEXT_PROTOTYPES // VBO
#include <wx/glcanvas.h>

#ifdef WIN32
// VBO extension definitions (glext.h)
#define GL_ARRAY_BUFFER_ARB 0x8892
#define GL_ELEMENT_ARRAY_BUFFER_ARB 0x8893
#define GL_STATIC_DRAW_ARB 0x88E4
typedef void (APIENTRY * PFNGLBINDBUFFERARBPROC) (GLenum target, GLuint buffer);
typedef void (APIENTRY * PFNGLDELETEBUFFERSARBPROC) (GLsizei n, const GLuint *buffers);
typedef void (APIENTRY * PFNGLGENBUFFERSARBPROC) (GLsizei n, GLuint *buffers);
typedef void (APIENTRY * PFNGLBUFFERDATAARBPROC) (GLenum target, int size, const GLvoid *data, GLenum usage);

PFNGLGENBUFFERSARBPROC glGenBuffersARB = NULL;
PFNGLBINDBUFFERARBPROC glBindBufferARB = NULL;
PFNGLBUFFERDATAARBPROC glBufferDataARB = NULL;
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB = NULL;
#endif

VertexTable::VertexTable(
	unsigned int vertexCount,
	bool useNormal,
	bool useColor,
	bool useTexture
	)
	:
	indices(NULL),
	vertices(NULL),
	normals(NULL),
	colors(NULL),
	texCoords(NULL),
	primitiveCount(0),
	primitiveType(GL_TRIANGLES),
	vertexCount(vertexCount),
	texture(NULL),
	deleteIndices(false)
{
	if( vertexCount != 0 )
	{
		wxASSERT(vertexCount < 65535); // else indices cannot be unsigned shorts
		vertices = new float[vertexCount * 3];

		if( useNormal )
			normals = new signed short[vertexCount * 3];

		if( useColor )
			colors = new unsigned char[vertexCount * 3];

		if( useTexture )
			texCoords = new float[vertexCount * 2];
	}
	memset(vboId, 0, 5 * sizeof(unsigned int));
}

VertexTable::VertexTable(
	unsigned short* indices,
	float* vertices,
	signed short* normals,
	unsigned char* colors,
	float* texCoords,
	unsigned int primitiveCount,
	unsigned int primitiveType,
	unsigned int vertexCount,
	Texture *texture
	)
	:
	indices(indices),
	vertices(vertices),
	normals(normals),
	colors(colors),
	texCoords(texCoords),
	primitiveCount(primitiveCount),
	primitiveType(primitiveType),
	vertexCount(vertexCount),
	texture(texture),
	deleteIndices(true)
{
}

VertexTable::~VertexTable()
{
	if(indices && deleteIndices)
		delete[] indices;
	if(vertices)
		delete[] vertices;
	if(normals)
		delete[] normals;
	if( colors )
		delete[] colors;
	if( texCoords )
		delete[] texCoords;
	if( vboId[0] )
		unLoadVBOs();
}

void VertexTable::clearPositions()
{
	memset(vertices, 0, vertexCount * 3 * sizeof(float));
}

void VertexTable::clearIndices()
{
	memset(indices, 0, primitiveCount * 3 * sizeof(unsigned short));
}

void VertexTable::setPosition(unsigned int index, const Vector &position)
{
	wxASSERT(index < vertexCount);
	vertices[index * 3	] = (float) position.x;
	vertices[index * 3 + 1	] = (float) position.y;
	vertices[index * 3 + 2	] = (float) position.z;
}

Vector VertexTable::getPosition(unsigned int index)
{
	wxASSERT(index < vertexCount && vertices);
	return Vector(
		vertices[index * 3	],
		vertices[index * 3 + 1	],
		vertices[index * 3 + 2	]);
}

void VertexTable::setNormal(unsigned int index, const Vector &normal)
{
	wxASSERT(index < vertexCount && normals);
	normals[index * 3	] = (signed short) normal.x;
	normals[index * 3 + 1	] = (signed short) normal.y;
	normals[index * 3 + 2	] = (signed short) normal.z;
}

void VertexTable::setColor(unsigned int index, const Color &color)
{//TODO: RGB or RGBA ?
	wxASSERT(index < vertexCount && colors);
	color.toRgb(&colors[index * 3]);
}

void VertexTable::setTextureCoordinates(unsigned int index, const Vector2 &coords)
{
	wxASSERT(index < vertexCount && texCoords);
	texCoords[index * 2	] = coords.u;
	texCoords[index * 2 + 1	] = coords.v;
}

void VertexTable::setPrimitiveCount(unsigned int count)
{
	primitiveCount = count;

	if( primitiveCount == 0 && indices != NULL && deleteIndices )
		delete[] indices;
	else
	{
		wxASSERT(indices == NULL); // resizing is not implemented
		indices = new unsigned short[primitiveCount * 3];
		deleteIndices = true; // The destructor will delete this array.
	}
}

void VertexTable::setPrimitive(unsigned int index, unsigned short i, unsigned short j, unsigned short k)
{
	wxASSERT(index < primitiveCount);
	indices[index * 3	] = i;
	indices[index * 3 + 1	] = j;
	indices[index * 3 + 2	] = k;
}

void VertexTable::setPrimitives(const VertexTable &table)
{
	primitiveCount = table.primitiveCount;
	indices = table.indices;
}

void VertexTable::draw(bool useTexture, bool useColor, bool useVBO)
{
	useVBO = false; // TODO: check if VBOs are available on the current machine, and if user wants them to be enabled
	if( useVBO && vboId[0] == 0 )
		loadVBOs(useTexture, useColor);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	if( useColor && colors )
		glEnableClientState(GL_COLOR_ARRAY);
	if( useTexture && texture )
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnable(GL_TEXTURE_2D);
		glBindTexture( GL_TEXTURE_2D, texture->getId() );
	}

	if(useVBO)
	{
		// Vertex coordinates
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[1]);
		glVertexPointer(3, GL_FLOAT, 0, NULL);

		// Normals
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[2]);
		glNormalPointer(GL_SHORT, 0, NULL);

		// Texture coordinates
		if( useTexture && texture )
		{
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[3]);
			glTexCoordPointer(2, GL_FLOAT, 0, NULL);
		}

		// Colors
		if( useColor && colors )
		{
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[4]);
			glColorPointer(3, GL_UNSIGNED_BYTE, 0, NULL);
		}

		glDrawElements(
			primitiveType,
			primitiveCount * 3,
			GL_UNSIGNED_SHORT,
			NULL);
	}
	else
	{
		// Vertex coordinates
		glVertexPointer(3, GL_FLOAT, 0, vertices);

		// Normals
		glNormalPointer(GL_SHORT, 0, normals);

		// Texture coordinates
		if( texture )
		{
			glTexCoordPointer(2, GL_FLOAT, 0, texCoords);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		}

		// Colors
		if( useColor && colors != NULL )
			glColorPointer(3, GL_UNSIGNED_BYTE, 0, NULL);

		glDrawElements(
			primitiveType,
			primitiveCount * 3,
			GL_UNSIGNED_SHORT,
			indices);
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	if( useColor && colors )
		glDisableClientState(GL_COLOR_ARRAY);
	if( useTexture && texture )
	{
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisable(GL_TEXTURE_2D);
	}
}

void VertexTable::loadVBOs(bool useTexture, bool useColor)
{
	// TODO: to be moved to another place
#ifdef WIN32
	glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) wglGetProcAddress("glGenBuffersARB");
	glBindBufferARB = (PFNGLBINDBUFFERARBPROC) wglGetProcAddress("glBindBufferARB");
	glBufferDataARB = (PFNGLBUFFERDATAARBPROC) wglGetProcAddress("glBufferDataARB");
	glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) wglGetProcAddress("glDeleteBuffersARB");
#endif

	glGenBuffersARB(5, vboId);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vboId[0]);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, primitiveCount * 3 * sizeof(unsigned short), indices, GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[1]);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexCount * 3 * sizeof(float), vertices, GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[2]);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexCount * 3 * sizeof(signed short), normals, GL_STATIC_DRAW_ARB);
	if( useTexture && texture )
	{
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[3]);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexCount * 2 * sizeof(float), texCoords, GL_STATIC_DRAW_ARB);
	}
	if( useColor && colors )
	{
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vboId[4]);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexCount * 3 * sizeof(unsigned char), colors, GL_STATIC_DRAW_ARB);
	}
}

void VertexTable::unLoadVBOs()
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
	glDeleteBuffersARB(5, vboId);
}
